@@ -2,63 +2,51 @@
 
 """Test case related stuff."""
 
-__metaclass__ = type
 __all__ = [
-    'attr',
-    'clone_test_with_new_id',
-    'ExpectedException',
-    'gather_details',
-    'run_test_with',
-    'skip',
-    'skipIf',
-    'skipUnless',
-    'TestCase',
-    ]
+    "ExpectedException",
+    "TestCase",
+    "attr",
+    "clone_test_with_new_id",
+    "gather_details",
+    "run_test_with",
+    "skip",
+    "skipIf",
+    "skipUnless",
+    "unique_text_generator",
+]
 
 import copy
 import functools
 import itertools
 import sys
-import types
+import unittest
+from typing import Any, Protocol, TypeVar, cast
+from unittest.case import SkipTest
 
-from extras import (
-    safe_hasattr,
-    try_import,
-    )
-import unittest2 as unittest
-
-from testtools import (
-    content,
-    )
-from testtools.compat import (
-    advance_iterator,
-    reraise,
-    )
+from testtools import content
+from testtools.compat import reraise
 from testtools.matchers import (
     Annotate,
     Contains,
-    Equals,
+    Is,
+    IsInstance,
+    Matcher,
     MatchesAll,
     MatchesException,
     MismatchError,
-    Is,
-    IsInstance,
     Not,
     Raises,
-    )
+)
+from testtools.matchers._basic import _FlippedEquals
 from testtools.monkey import patch
-from testtools.runtest import RunTest
+from testtools.runtest import (
+    MultipleExceptions,
+    RunTest,
+)
 from testtools.testresult import (
     ExtendedToOriginalDecorator,
     TestResult,
-    )
-
-wraps = try_import('functools.wraps')
-
-class TestSkipped(Exception):
-    """Raised within TestCase.run() when a test is skipped."""
-TestSkipped = try_import('unittest.case.SkipTest', TestSkipped)
-TestSkipped = try_import('unittest2.case.SkipTest', TestSkipped)
+)
 
 
 class _UnexpectedSuccess(Exception):
@@ -67,21 +55,14 @@ class _UnexpectedSuccess(Exception):
     Note that this exception is private plumbing in testtools' testcase
     module.
     """
-_UnexpectedSuccess = try_import(
-    'unittest.case._UnexpectedSuccess', _UnexpectedSuccess)
-_UnexpectedSuccess = try_import(
-    'unittest2.case._UnexpectedSuccess', _UnexpectedSuccess)
+
 
 class _ExpectedFailure(Exception):
-    """An expected failure occured.
+    """An expected failure occurred.
 
     Note that this exception is private plumbing in testtools' testcase
     module.
     """
-_ExpectedFailure = try_import(
-    'unittest.case._ExpectedFailure', _ExpectedFailure)
-_ExpectedFailure = try_import(
-    'unittest2.case._ExpectedFailure', _ExpectedFailure)
 
 
 # Copied from unittest before python 3.4 release. Used to maintain
@@ -95,6 +76,7 @@ def _expectedFailure(func):
         except Exception:
             raise _ExpectedFailure(sys.exc_info())
         raise _UnexpectedSuccess
+
     return wrapper
 
 
@@ -123,20 +105,23 @@ def run_test_with(test_runner, **kwargs):
     :return: A decorator to be used for marking a test as needing a special
         runner.
     """
+
     def decorator(function):
         # Set an attribute on 'function' which will inform TestCase how to
         # make the runner.
         def _run_test_with(case, handlers=None, last_resort=None):
             try:
                 return test_runner(
-                    case, handlers=handlers, last_resort=last_resort,
-                    **kwargs)
+                    case, handlers=handlers, last_resort=last_resort, **kwargs
+                )
             except TypeError:
                 # Backwards compat: if we can't call the constructor
                 # with last_resort, try without that.
                 return test_runner(case, handlers=handlers, **kwargs)
+
         function._run_test_with = _run_test_with
         return function
+
     return decorator
 
 
@@ -152,13 +137,19 @@ def _copy_content(content_object):
         ``content_object`` and a non-volatile copy of its content.
     """
     content_bytes = list(content_object.iter_bytes())
-    content_callback = lambda: content_bytes
+
+    def content_callback():
+        return content_bytes
+
     return content.Content(content_object.content_type, content_callback)
 
 
 def gather_details(source_dict, target_dict):
     """Merge the details from ``source_dict`` into ``target_dict``.
 
+    ``gather_details`` evaluates all details in ``source_dict``. Do not use it
+    if the details are not ready to be evaluated.
+
     :param source_dict: A dictionary of details will be gathered.
     :param target_dict: A dictionary into which details will be gathered.
     """
@@ -166,11 +157,67 @@ def gather_details(source_dict, target_dict):
         new_name = name
         disambiguator = itertools.count(1)
         while new_name in target_dict:
-            new_name = '%s-%d' % (name, advance_iterator(disambiguator))
+            new_name = f"{name}-{next(disambiguator)}"
         name = new_name
         target_dict[name] = _copy_content(content_object)
 
 
+# Circular import: fixtures imports gather_details from here, we import
+# fixtures, leading to gather_details not being available and fixtures being
+# unable to import it.
+try:
+    import fixtures
+except ImportError:
+    fixtures = None  # type: ignore
+
+
+class UseFixtureProtocol(Protocol):
+    def setUp(self) -> Any: ...
+    def cleanUp(self) -> Any: ...
+    def getDetails(self) -> dict: ...
+
+
+UseFixtureT = TypeVar("UseFixtureT", bound=UseFixtureProtocol)
+
+
+def _mods(i, mod):
+    (q, r) = divmod(i, mod)
+    while True:
+        yield r
+        if not q:
+            break
+        (q, r) = divmod(q, mod)
+
+
+def _unique_text(base_cp, cp_range, index):
+    s = ""
+    for m in _mods(index, cp_range):
+        s += chr(base_cp + m)
+    return s
+
+
+def unique_text_generator(prefix):
+    """Generates unique text values.
+
+    Generates text values that are unique. Use this when you need arbitrary
+    text in your test, or as a helper for custom anonymous factory methods.
+
+    :param prefix: The prefix for text.
+    :return: text that looks like '<prefix>-<text_with_unicode>'.
+    :rtype: str
+    """
+    # 0x1e00 is the start of a range of glyphs that are easy to see are
+    # unicode since they've got circles and dots and other diacriticals.
+    # 0x1eff is the end of the range of these diacritical chars.
+    BASE_CP = 0x1E00
+    CP_RANGE = 0x1F00 - BASE_CP
+    index = 0
+    while True:
+        unique_text = _unique_text(BASE_CP, CP_RANGE, index)
+        yield f"{prefix}-{unique_text}"
+        index = index + 1
+
+
 class TestCase(unittest.TestCase):
     """Extensions to the basic TestCase.
 
@@ -184,7 +231,7 @@ class TestCase(unittest.TestCase):
         and an optional list of exception handlers.
     """
 
-    skipException = TestSkipped
+    skipException = SkipTest
 
     run_tests_with = RunTest
 
@@ -197,24 +244,14 @@ class TestCase(unittest.TestCase):
             when run() is invoked, so will be fresh each time. Overrides
             ``TestCase.run_tests_with`` if given.
         """
-        runTest = kwargs.pop('runTest', None)
-        super(TestCase, self).__init__(*args, **kwargs)
-        self._cleanups = []
-        self._unique_id_gen = itertools.count(1)
-        # Generators to ensure unique traceback ids.  Maps traceback label to
-        # iterators.
-        self._traceback_id_gens = {}
-        self.__setup_called = False
-        self.__teardown_called = False
-        # __details is lazy-initialized so that a constructed-but-not-run
-        # TestCase is safe to use with clone_test_with_new_id.
-        self.__details = None
+        runTest = kwargs.pop("runTest", None)
+        super().__init__(*args, **kwargs)
+        self._reset()
         test_method = self._get_test_method()
         if runTest is None:
-            runTest = getattr(
-                test_method, '_run_test_with', self.run_tests_with)
+            runTest = getattr(test_method, "_run_test_with", self.run_tests_with)
         self.__RunTest = runTest
-        if getattr(test_method, '__unittest_expecting_failure__', False):
+        if getattr(test_method, "__unittest_expecting_failure__", False):
             setattr(self, self._testMethodName, _expectedFailure(test_method))
         # Used internally for onException processing - used to gather extra
         # data from exceptions.
@@ -226,17 +263,36 @@ class TestCase(unittest.TestCase):
             (_ExpectedFailure, self._report_expected_failure),
             (_UnexpectedSuccess, self._report_unexpected_success),
             (Exception, self._report_error),
-            ]
+        ]
+
+    def _reset(self):
+        """Reset the test case as if it had never been run."""
+        self._cleanups = []
+        self._unique_id_gen = itertools.count(1)
+        # Generators to ensure unique traceback ids.  Maps traceback label to
+        # iterators.
+        self._traceback_id_gens = {}
+        self.__setup_called = False
+        self.__teardown_called = False
+        # __details is lazy-initialized so that a constructed-but-not-run
+        # TestCase is safe to use with clone_test_with_new_id.
+        self.__details = None
 
     def __eq__(self, other):
-        eq = getattr(unittest.TestCase, '__eq__', None)
-        if eq is not None and not unittest.TestCase.__eq__(self, other):
-            return False
-        return self.__dict__ == other.__dict__
+        eq = getattr(unittest.TestCase, "__eq__", None)
+        if eq is not None:
+            eq_ = unittest.TestCase.__eq__(self, other)
+            if eq_ is NotImplemented or not eq_:
+                return False
+        return self.__dict__ == getattr(other, "__dict__", None)
+
+    # We need to explicitly set this since we're overriding __eq__
+    # https://docs.python.org/3/reference/datamodel.html#object.__hash__
+    __hash__ = unittest.TestCase.__hash__
 
     def __repr__(self):
         # We add id to the repr because it makes testing testtools easier.
-        return "<%s id=0x%0x>" % (self.id(), id(self))
+        return f"<{self.id()} id=0x{id(self):0x}>"
 
     def addDetail(self, name, content_object):
         """Add a detail to be reported with this test's outcome.
@@ -289,18 +345,14 @@ class TestCase(unittest.TestCase):
         """
         raise self.skipException(reason)
 
-    # skipTest is how python2.7 spells this. Sometime in the future
-    # This should be given a deprecation decorator - RBC 20100611.
-    skip = skipTest
-
     def _formatTypes(self, classOrIterable):
         """Format a class or a bunch of classes for display in an error."""
-        className = getattr(classOrIterable, '__name__', None)
+        className = getattr(classOrIterable, "__name__", None)
         if className is None:
-            className = ', '.join(klass.__name__ for klass in classOrIterable)
+            className = ", ".join(klass.__name__ for klass in classOrIterable)
         return className
 
-    def addCleanup(self, function, *arguments, **keywordArguments):
+    def addCleanup(self, function, /, *arguments, **keywordArguments):
         """Add a cleanup function to be called after tearDown.
 
         Functions added with addCleanup will be called in reverse order of
@@ -328,32 +380,30 @@ class TestCase(unittest.TestCase):
         the exception has caused.
 
         Handlers are called in first-added, first-called order, and if they
-        raise an exception, that will propogate out of the test running
+        raise an exception, that will propagate out of the test running
         machinery, halting test processing. As a result, do not call code that
         may unreasonably fail.
         """
         self.__exception_handlers.append(handler)
 
     def _add_reason(self, reason):
-        self.addDetail('reason', content.text_content(reason))
+        self.addDetail("reason", content.text_content(reason))
 
-    def assertEqual(self, expected, observed, message=''):
+    def assertEqual(self, expected, observed, message=""):
         """Assert that 'expected' is equal to 'observed'.
 
         :param expected: The expected value.
         :param observed: The observed value.
         :param message: An optional message to include in the error.
         """
-        matcher = Equals(expected)
+        matcher = _FlippedEquals(expected)
         self.assertThat(observed, matcher, message)
 
-    failUnlessEqual = assertEquals = assertEqual
-
-    def assertIn(self, needle, haystack, message=''):
+    def assertIn(self, needle, haystack, message=""):
         """Assert that needle is in haystack."""
         self.assertThat(haystack, Contains(needle), message)
 
-    def assertIsNone(self, observed, message=''):
+    def assertIsNone(self, observed, message=""):
         """Assert that 'observed' is equal to None.
 
         :param observed: The observed value.
@@ -362,7 +412,7 @@ class TestCase(unittest.TestCase):
         matcher = Is(None)
         self.assertThat(observed, matcher, message)
 
-    def assertIsNotNone(self, observed, message=''):
+    def assertIsNotNone(self, observed, message=""):
         """Assert that 'observed' is not equal to None.
 
         :param observed: The observed value.
@@ -371,7 +421,7 @@ class TestCase(unittest.TestCase):
         matcher = Not(Is(None))
         self.assertThat(observed, matcher, message)
 
-    def assertIs(self, expected, observed, message=''):
+    def assertIs(self, expected, observed, message=""):
         """Assert that 'expected' is 'observed'.
 
         :param expected: The expected value.
@@ -381,12 +431,12 @@ class TestCase(unittest.TestCase):
         matcher = Is(expected)
         self.assertThat(observed, matcher, message)
 
-    def assertIsNot(self, expected, observed, message=''):
+    def assertIsNot(self, expected, observed, message=""):
         """Assert that 'expected' is not 'observed'."""
         matcher = Not(Is(expected))
         self.assertThat(observed, matcher, message)
 
-    def assertNotIn(self, needle, haystack, message=''):
+    def assertNotIn(self, needle, haystack, message=""):
         """Assert that needle is not in haystack."""
         matcher = Not(Contains(needle))
         self.assertThat(haystack, matcher, message)
@@ -398,30 +448,53 @@ class TestCase(unittest.TestCase):
             matcher = IsInstance(klass)
         self.assertThat(obj, matcher, msg)
 
-    def assertRaises(self, excClass, callableObj, *args, **kwargs):
-        """Fail unless an exception of class excClass is thrown
-           by callableObj when invoked with arguments args and keyword
-           arguments kwargs. If a different type of exception is
-           thrown, it will not be caught, and the test case will be
-           deemed to have suffered an error, exactly as for an
-           unexpected exception.
+    def assertRaises(self, expected_exception, callable=None, *args, **kwargs):
+        """Fail unless an exception of class expected_exception is thrown
+        by callable when invoked with arguments args and keyword
+        arguments kwargs. If a different type of exception is
+        thrown, it will not be caught, and the test case will be
+        deemed to have suffered an error, exactly as for an
+        unexpected exception.
+
+        If called with the callable omitted, will return a
+        context object used like this::
+
+            with self.assertRaises(SomeException):
+                do_something()
+
+        The context manager keeps a reference to the exception as
+        the 'exception' attribute. This allows you to inspect the
+        exception after the assertion::
+
+            with self.assertRaises(SomeException) as cm:
+                do_something()
+            the_exception = cm.exception
+            self.assertEqual(the_exception.error_code, 3)
         """
-        class ReRaiseOtherTypes(object):
+        # If callable is None, we're being used as a context manager
+        if callable is None:
+            return _AssertRaisesContext(expected_exception, self, msg=kwargs.get("msg"))
+
+        class ReRaiseOtherTypes:
             def match(self, matchee):
-                if not issubclass(matchee[0], excClass):
+                if not issubclass(matchee[0], expected_exception):
                     reraise(*matchee)
-        class CaptureMatchee(object):
+
+        class CaptureMatchee:
             def match(self, matchee):
                 self.matchee = matchee[1]
+
         capture = CaptureMatchee()
-        matcher = Raises(MatchesAll(ReRaiseOtherTypes(),
-                MatchesException(excClass), capture))
-        our_callable = Nullary(callableObj, *args, **kwargs)
+        matcher = Raises(
+            MatchesAll(
+                ReRaiseOtherTypes(), MatchesException(expected_exception), capture
+            )
+        )
+        our_callable = Nullary(callable, *args, **kwargs)
         self.assertThat(our_callable, matcher)
         return capture.matchee
-    failUnlessRaises = assertRaises
 
-    def assertThat(self, matchee, matcher, message='', verbose=False):
+    def assertThat(self, matchee, matcher, message="", verbose=False):
         """Assert that matchee is matched by matcher.
 
         :param matchee: An object to match with matcher.
@@ -449,21 +522,22 @@ class TestCase(unittest.TestCase):
         full_name = name
         suffix = 1
         while full_name in existing_details:
-            full_name = "%s-%d" % (name, suffix)
+            full_name = f"{name}-{suffix}"
             suffix += 1
         self.addDetail(full_name, content_object)
 
-    def expectThat(self, matchee, matcher, message='', verbose=False):
+    def expectThat(self, matchee, matcher, message="", verbose=False):
         """Check that matchee is matched by matcher, but delay the assertion failure.
 
         This method behaves similarly to ``assertThat``, except that a failed
-        match does not exit the test immediately. The rest of the test code will
-        continue to run, and the test will be marked as failing after the test
-        has finished.
+        match does not exit the test immediately. The rest of the test code
+        will continue to run, and the test will be marked as failing after the
+        test has finished.
 
         :param matchee: An object to match with matcher.
         :param matcher: An object meeting the testtools.Matcher protocol.
         :param message: If specified, show this message with any failed match.
+
         """
         mismatch_error = self._matchHelper(matchee, matcher, message, verbose)
 
@@ -472,7 +546,7 @@ class TestCase(unittest.TestCase):
                 "Failed expectation",
                 content.StacktraceContent(
                     postfix_content="MismatchError: " + str(mismatch_error)
-                )
+                ),
             )
             self.force_failure = True
 
@@ -481,7 +555,7 @@ class TestCase(unittest.TestCase):
         mismatch = matcher.match(matchee)
         if not mismatch:
             return
-        for (name, value) in mismatch.get_details().items():
+        for name, value in mismatch.get_details().items():
             self.addDetailUniqueName(name, value)
         return MismatchError(matchee, matcher, mismatch, verbose)
 
@@ -527,7 +601,7 @@ class TestCase(unittest.TestCase):
         Use this when you need an arbitrary integer in your test, or as a
         helper for custom anonymous factory methods.
         """
-        return advance_iterator(self._unique_id_gen)
+        return next(self._unique_id_gen)
 
     def getUniqueString(self, prefix=None):
         """Get a string unique to this test.
@@ -542,15 +616,18 @@ class TestCase(unittest.TestCase):
         """
         if prefix is None:
             prefix = self.id()
-        return '%s-%d' % (prefix, self.getUniqueInteger())
+        return f"{prefix}-{self.getUniqueInteger()}"
 
-    def onException(self, exc_info, tb_label='traceback'):
-        """Called when an exception propogates from test code.
+    def onException(self, exc_info, tb_label="traceback"):
+        """Called when an exception propagates from test code.
 
         :seealso addOnException:
         """
         if exc_info[0] not in [
-            TestSkipped, _UnexpectedSuccess, _ExpectedFailure]:
+            self.skipException,
+            _UnexpectedSuccess,
+            _ExpectedFailure,
+        ]:
             self._report_traceback(exc_info, tb_label=tb_label)
         for handler in self.__exception_handlers:
             handler(exc_info)
@@ -576,25 +653,33 @@ class TestCase(unittest.TestCase):
         self._add_reason(reason)
         result.addSkip(self, details=self.getDetails())
 
-    def _report_traceback(self, exc_info, tb_label='traceback'):
-        id_gen = self._traceback_id_gens.setdefault(
-            tb_label, itertools.count(0))
+    def _report_traceback(self, exc_info, tb_label="traceback"):
+        id_gen = self._traceback_id_gens.setdefault(tb_label, itertools.count(0))
         while True:
-            tb_id = advance_iterator(id_gen)
+            tb_id = next(id_gen)
             if tb_id:
-                tb_label = '%s-%d' % (tb_label, tb_id)
+                tb_label = f"{tb_label}-{tb_id}"
             if tb_label not in self.getDetails():
                 break
-        self.addDetail(tb_label, content.TracebackContent(exc_info, self))
+        self.addDetail(
+            tb_label,
+            content.TracebackContent(
+                exc_info,
+                self,
+                capture_locals=getattr(self, "__testtools_tb_locals__", False),
+            ),
+        )
 
     @staticmethod
     def _report_unexpected_success(self, result, err):
         result.addUnexpectedSuccess(self, details=self.getDetails())
 
     def run(self, result=None):
+        self._reset()
         try:
             run_test = self.__RunTest(
-                self, self.exception_handlers, last_resort=self._report_error)
+                self, self.exception_handlers, last_resort=self._report_error
+            )
         except TypeError:
             # Backwards compat: if we can't call the constructor
             # with last_resort, try without that.
@@ -611,12 +696,11 @@ class TestCase(unittest.TestCase):
         ret = self.setUp()
         if not self.__setup_called:
             raise ValueError(
-                "In File: %s\n"
+                f"In File: {sys.modules[self.__class__.__module__].__file__}\n"
                 "TestCase.setUp was not called. Have you upcalled all the "
                 "way up the hierarchy from your setUp? e.g. Call "
-                "super(%s, self).setUp() from your setUp()."
-                % (sys.modules[self.__class__.__module__].__file__,
-                   self.__class__.__name__))
+                f"super({self.__class__.__name__}, self).setUp() from your setUp()."
+            )
         return ret
 
     def _run_teardown(self, result):
@@ -629,17 +713,27 @@ class TestCase(unittest.TestCase):
         ret = self.tearDown()
         if not self.__teardown_called:
             raise ValueError(
-                "In File: %s\n"
+                f"In File: {sys.modules[self.__class__.__module__].__file__}\n"
                 "TestCase.tearDown was not called. Have you upcalled all the "
                 "way up the hierarchy from your tearDown? e.g. Call "
-                "super(%s, self).tearDown() from your tearDown()."
-                % (sys.modules[self.__class__.__module__].__file__,
-                   self.__class__.__name__))
+                f"super({self.__class__.__name__}, self).tearDown() "
+                "from your tearDown()."
+            )
         return ret
 
     def _get_test_method(self):
-        method_name = getattr(self, '_testMethodName')
-        return getattr(self, method_name)
+        method_name = getattr(self, "_testMethodName")
+        try:
+            m = getattr(self, method_name)
+        except AttributeError:
+            if method_name != "runTest":
+                # We allow instantiation with no explicit method name
+                # but not an *incorrect* or missing method name.
+                raise ValueError(
+                    f"no such test method in {self.__class__}: {method_name}"
+                )
+        else:
+            return m
 
     def _run_test_method(self, result):
         """Run the test method for this test.
@@ -649,7 +743,7 @@ class TestCase(unittest.TestCase):
         """
         return self._get_test_method()()
 
-    def useFixture(self, fixture):
+    def useFixture(self, fixture: UseFixtureT) -> UseFixtureT:
         """Use fixture in a test case.
 
         The fixture will be setUp, and self.addCleanup(fixture.cleanUp) called.
@@ -660,11 +754,21 @@ class TestCase(unittest.TestCase):
         """
         try:
             fixture.setUp()
-        except:
+        except MultipleExceptions as e:
+            if fixtures is not None and e.args[-1][0] is fixtures.fixture.SetupError:
+                gather_details(e.args[-1][1].args[0], self.getDetails())
+            raise
+        except BaseException:
             exc_info = sys.exc_info()
             try:
-                gather_details(fixture.getDetails(), self.getDetails())
-            except:
+                # fixture._details is not available if using the newer
+                # _setUp() API in Fixtures because it already cleaned up
+                # the fixture.  Ideally this whole try/except is not
+                # really needed any more, however, we keep this code to
+                # remain compatible with the older setUp().
+                if hasattr(fixture, "_details") and fixture._details is not None:
+                    gather_details(fixture.getDetails(), self.getDetails())
+            except BaseException:
                 # Report the setUp exception, then raise the error during
                 # gather_details.
                 self._report_traceback(exc_info)
@@ -675,34 +779,33 @@ class TestCase(unittest.TestCase):
                 reraise(*exc_info)
         else:
             self.addCleanup(fixture.cleanUp)
-            self.addCleanup(
-                gather_details, fixture.getDetails(), self.getDetails())
+            self.addCleanup(gather_details, fixture.getDetails(), self.getDetails())
             return fixture
 
     def setUp(self):
-        super(TestCase, self).setUp()
+        super().setUp()
         if self.__setup_called:
             raise ValueError(
-                "In File: %s\n"
+                f"In File: {sys.modules[self.__class__.__module__].__file__}\n"
                 "TestCase.setUp was already called. Do not explicitly call "
                 "setUp from your tests. In your own setUp, use super to call "
                 "the base setUp."
-                % (sys.modules[self.__class__.__module__].__file__,))
+            )
         self.__setup_called = True
 
     def tearDown(self):
-        super(TestCase, self).tearDown()
+        super().tearDown()
         if self.__teardown_called:
             raise ValueError(
-                "In File: %s\n"
+                f"In File: {sys.modules[self.__class__.__module__].__file__}\n"
                 "TestCase.tearDown was already called. Do not explicitly call "
                 "tearDown from your tests. In your own tearDown, use super to "
                 "call the base tearDown."
-                % (sys.modules[self.__class__.__module__].__file__,))
+            )
         self.__teardown_called = True
 
 
-class PlaceHolder(object):
+class PlaceHolder:
     """A placeholder test.
 
     `PlaceHolder` implements much of the same interface as TestCase and is
@@ -711,8 +814,16 @@ class PlaceHolder(object):
 
     failureException = None
 
-    def __init__(self, test_id, short_description=None, details=None,
-        outcome='addSuccess', error=None, tags=None, timestamps=(None, None)):
+    def __init__(
+        self,
+        test_id,
+        short_description=None,
+        details=None,
+        outcome="addSuccess",
+        error=None,
+        tags=None,
+        timestamps=(None, None),
+    ):
         """Construct a `PlaceHolder`.
 
         :param test_id: The id of the placeholder test.
@@ -729,7 +840,7 @@ class PlaceHolder(object):
         self._details = details or {}
         self._outcome = outcome
         if error is not None:
-            self._details['traceback'] = content.TracebackContent(error, self)
+            self._details["traceback"] = content.TracebackContent(error, self)
         tags = tags or frozenset()
         self._tags = frozenset(tags)
         self._timestamps = timestamps
@@ -741,10 +852,11 @@ class PlaceHolder(object):
         internal = [self._outcome, self._test_id, self._details]
         if self._short_description is not None:
             internal.append(self._short_description)
-        return "<%s.%s(%s)>" % (
+        return "<{}.{}({})>".format(
             self.__class__.__module__,
             self.__class__.__name__,
-            ", ".join(map(repr, internal)))
+            ", ".join(map(repr, internal)),
+        )
 
     def __str__(self):
         return self.id()
@@ -794,8 +906,13 @@ def ErrorHolder(test_id, error, short_description=None, details=None):
     :param short_description: An optional short description of the test.
     :param details: Outcome details as accepted by addSuccess etc.
     """
-    return PlaceHolder(test_id, short_description=short_description,
-        details=details, outcome='addError', error=error)
+    return PlaceHolder(
+        test_id,
+        short_description=short_description,
+        details=details,
+        outcome="addError",
+        error=error,
+    )
 
 
 def _clone_test_id_callback(test, callback):
@@ -829,15 +946,17 @@ def attr(*args):
     :return: A callable that when applied to a WithAttributes will
         alter its id to enumerate the added attributes.
     """
+
     def decorate(fn):
-        if not safe_hasattr(fn, '__testtools_attrs'):
+        if not hasattr(fn, "__testtools_attrs"):
             fn.__testtools_attrs = set()
         fn.__testtools_attrs.update(args)
         return fn
+
     return decorate
 
 
-class WithAttributes(object):
+class WithAttributes:
     """A mix-in class for modifying test id by attributes.
 
     e.g.
@@ -849,15 +968,20 @@ class WithAttributes(object):
     testtools.testcase.MyTest/test_bar[foo]
     """
 
-    def id(self):
-        orig = super(WithAttributes, self).id()
+    _get_test_method: Any  # Provided by the class we're mixed with
+
+    def id(self) -> str:
+        orig = cast(str, super().id())  # type: ignore[misc]
         # Depends on testtools.TestCase._get_test_method, be nice to support
         # plain unittest.
         fn = self._get_test_method()
-        attributes = getattr(fn, '__testtools_attrs', None)
+        attributes = cast(str, getattr(fn, "__testtools_attrs", None))
         if not attributes:
             return orig
-        return orig + '[' + ','.join(sorted(attributes)) + ']'
+        return orig + "[" + ",".join(sorted(attributes)) + "]"
+
+
+class_types = (type,)
 
 
 def skip(reason):
@@ -867,21 +991,24 @@ def skip(reason):
     unit tests in order to migrate to python 2.7, which provides the
     @unittest.skip decorator.
     """
+
     def decorator(test_item):
+        if not isinstance(test_item, class_types):
+
+            @functools.wraps(test_item)
+            def skip_wrapper(*args, **kwargs):
+                raise TestCase.skipException(reason)
+
+            test_item = skip_wrapper
+
         # This attribute signals to RunTest._run_core that the entire test
         # must be skipped - including setUp and tearDown. This makes us
         # compatible with testtools.skip* functions, which set the same
         # attributes.
         test_item.__unittest_skip__ = True
         test_item.__unittest_skip_why__ = reason
-        if wraps is not None:
-            @wraps(test_item)
-            def skip_wrapper(*args, **kwargs):
-                raise TestCase.skipException(reason)
-        else:
-            def skip_wrapper(test_item):
-                test_item.skip(reason)
-        return skip_wrapper
+        return test_item
+
     return decorator
 
 
@@ -889,8 +1016,10 @@ def skipIf(condition, reason):
     """A decorator to skip a test if the condition is true."""
     if condition:
         return skip(reason)
+
     def _id(obj):
         return obj
+
     return _id
 
 
@@ -898,11 +1027,58 @@ def skipUnless(condition, reason):
     """A decorator to skip a test unless the condition is true."""
     if not condition:
         return skip(reason)
+
     def _id(obj):
         return obj
+
     return _id
 
 
+class _AssertRaisesContext:
+    """A context manager to handle expected exceptions for assertRaises.
+
+    This provides compatibility with unittest's assertRaises context manager.
+    """
+
+    def __init__(self, expected, test_case, msg=None):
+        """Construct an `_AssertRaisesContext`.
+
+        :param expected: The type of exception to expect.
+        :param test_case: The TestCase instance using this context.
+        :param msg: An optional message explaining the failure.
+        """
+        self.expected = expected
+        self.test_case = test_case
+        self.msg = msg
+        self.exception = None
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        if exc_type is None:
+            try:
+                if isinstance(self.expected, tuple):
+                    exc_name = "({})".format(
+                        ", ".join(e.__name__ for e in self.expected)
+                    )
+                else:
+                    exc_name = self.expected.__name__
+            except AttributeError:
+                exc_name = str(self.expected)
+            if self.msg:
+                error_msg = f"{exc_name} not raised : {self.msg}"
+            else:
+                error_msg = f"{exc_name} not raised"
+            raise self.test_case.failureException(error_msg)
+        if not issubclass(exc_type, self.expected):
+            # let unexpected exceptions pass through
+            return False
+        # store exception for later retrieval
+        self.exception = exc_value
+        return True
+
+
 class ExpectedException:
     """A context manager to handle expected exceptions.
 
@@ -933,23 +1109,26 @@ class ExpectedException:
 
     def __exit__(self, exc_type, exc_value, traceback):
         if exc_type is None:
-            error_msg = '%s not raised.' % self.exc_type.__name__
+            error_msg = f"{self.exc_type.__name__} not raised."
             if self.msg:
-                error_msg = error_msg + ' : ' + self.msg
+                error_msg = error_msg + " : " + self.msg
             raise AssertionError(error_msg)
         if exc_type != self.exc_type:
             return False
         if self.value_re:
-            matcher = MatchesException(self.exc_type, self.value_re)
+            exception_matcher = MatchesException(self.exc_type, self.value_re)
+            matcher: Matcher | Annotate
             if self.msg:
-                matcher = Annotate(self.msg, matcher)
+                matcher = Annotate(self.msg, exception_matcher)
+            else:
+                matcher = exception_matcher
             mismatch = matcher.match((exc_type, exc_value, traceback))
             if mismatch:
                 raise AssertionError(mismatch.describe())
         return True
 
 
-class Nullary(object):
+class Nullary:
     """Turn a callable into a nullary callable.
 
     The advantage of this over ``lambda: f(*args, **kwargs)`` is that it
@@ -968,7 +1147,7 @@ class Nullary(object):
         return repr(self._callable_object)
 
 
-class DecorateTestCaseResult(object):
+class DecorateTestCaseResult:
     """Decorate a TestCase and permit customisation of the result for runs."""
 
     def __init__(self, case, callout, before_run=None, after_run=None):
@@ -1011,7 +1190,7 @@ class DecorateTestCaseResult(object):
         delattr(self.decorated, name)
 
     def __setattr__(self, name, value):
-        if name in ('decorated', 'callout', 'before_run', 'after_run'):
+        if name in ("decorated", "callout", "before_run", "after_run"):
             self.__dict__[name] = value
             return
         setattr(self.decorated, name, value)
